import re
import os
import sys
import pathlib
import inspect

# CamelCase to snake_case
# https://stackoverflow.com/questions/1175208/elegant-python-function-to-convert-camelcase-to-snake-case
r1 = re.compile(r'(.)([A-Z][a-z]+)')
r2 = re.compile(r'__([A-Z])')
r3 = re.compile(r'([a-z0-9])([A-Z])')


def to_snake_case(name):
    name = r1.sub(r'\1_\2', name)
    name = r2.sub(r'_\1', name)
    name = r3.sub(r'\1_\2', name)
    return name.lower()


def component(**kwargs):
    def _(c):
        if not hasattr(c, '__data__'):
            class_name = to_snake_case(c.__name__)
            class_file_path = os.path.abspath((sys.modules[c.__module__].__file__))
            class_dir_path = str(pathlib.Path(class_file_path).parent.resolve())

            class_members = {t[0]: t[1] for t in inspect.getmembers(c)}
            init_members = {t[0]: t[1] for t in inspect.getmembers(class_members['__init__'])}
            init_annotations = init_members['__annotations__'] if '__annotations__' in init_members else {}

            extends = kwargs['extends'] if 'extends' in kwargs else None
            rel_template_path = kwargs['template_path'] if 'template_path' in kwargs else class_name + '.xml'
            template_path = os.path.join(class_dir_path, rel_template_path)
            components = kwargs['components'] if 'components' in kwargs else []
            components = {c.__data__['tag']: c for c in components}
            template = kwargs['template'] if 'template' in kwargs else None
            tag = kwargs['tag'] if 'tag' in kwargs else None
            defaults = kwargs['defaults'] if 'defaults' in kwargs else {}

            interfaces = {k[10:]: kwargs[k] for k in kwargs if k.startswith('interface_')}
            if len(interfaces) > 0:
                print("[deprecated] in "
                      + class_name
                      + ", interface_mjc and interface_hw should be "
                        "replace with defaults={'interface_mjc':xxx, 'interface_hw':xx}")
                defaults = {**{k: interfaces[k[10:]] for k in kwargs if k.startswith('interface_')}, **defaults}

            probe = kwargs['probe'] if 'probe' in kwargs else None
            if probe is not None:
                print("[deprecated] in "
                      + class_name
                      + ", interface_mjc and interface_hw should be "
                        "replace with defaults={'probe':xxx'}")
                defaults = {**{'probe': probe}, **defaults}

            setattr(c, '__data__', {
                'tag': tag if tag is not None else class_name,
                'class_name': class_name,  # the name of the factory class, in snake case
                'class_file_path': class_file_path,  # the abs path of the FILE where the class is declared
                'class_dir_path': class_dir_path,  # the abs path of the DIRECTORY where the class is declared
                'template': template,
                'template_path': template_path,  # the abs path of the .xml template
                'defaults': defaults,  # component config defaults
                # the dict of components' classes indexed by class name
                # e.g.:  {'ur5': <class 'yarok.components.ur5.ur5.UR5'>, ...}
                'components': components,
                'extends': extends,
                # the class constructor annotations
                # format: dict({ comp. name : class }) e.g.
                # e.g.:  {'mani': <class 'yarok.components.ur5.ur5.UR5'>, ...}
                'annotations': init_annotations,

                # interfaces's classes indexed environment "name"
                # e.g.: {'mjc': <class 'yarok.components.ur5.ur5.UR5InterfaceMJC'>}
                # 'interfaces': interfaces,

                # probe. a fn that returns an object key:value
                # the value is plotted or logged depending on the type.
                # 'probe': probe
            })
        return c

    return _



from yarok import component, interface


@component(
    tag='tumble-tower',
    defaults={
      'rows': 8
    },
    # language=xml
    template="""
        <mujoco>
            <asset>
                <material name="wood" rgba="0.5 0.4 0.3 1" specular="0.95"/>
            </asset>
            <default>
                <default class='tt-block'>
                    <geom type='box'
                     material='wood'
                            mass='0.1'
                            size="0.04 0.12 0.028"/>
                    
                </default>
            </default>
            <worldbody>
                <for each='range(rows)' as='z'>
                    <for each='range(3)' as='x'>
                        <body pos="
                                ${0.5 + 0.082*x if z % 2 == 0 else 0.58} 
                                ${0.48 if z % 2 == 0 else 0.4 + 0.082*x}
                                ${0.061*z}" 
                            euler="0 0 ${0 if z % 2 == 0 else 1.57}">
                            <freejoint/>
                            <geom class='tt-block'/>
                        </body>
                    </for>
                </for>
                <!--
            
                <body>
                    <geom class='tt-block-left'/>
                    <geom class='tt-block-center'/>
                    <geom class='tt-block-right'/>
                </body>
                
                <body euler='0 0 1.57' pos='0 0 0.061'>
                    <geom class='tt-block-left'/>
                    <geom class='tt-block-center'/>
                    <geom class='tt-block-right'/>
                </body>
                
                <body pos='0 0 0.121'>
                    <geom class='tt-block-left'/>
                    <geom class='tt-block-center'/>
                    <geom class='tt-block-right'/>
                </body>
                
                <body euler='0 0 1.57' pos='0 0 0.181'>
                    <geom class='tt-block-left'/>
                    <geom class='tt-block-center'/>
                    <geom class='tt-block-right'/>
                </body>
                
                <body pos='0 0 0.241'>
                    <geom class='tt-block-left'/>
                    <geom class='tt-block-center'/>
                    <geom class='tt-block-right'/>
                </body>
                
                <body euler='0 0 1.57' pos='0 0 0.301'>
                    <geom class='tt-block-left'/>
                    <geom class='tt-block-center'/>
                    <geom class='tt-block-right'/>
                </body>
                
                <body pos='0 0 0.361'>
                    <geom class='tt-block-left'/>
                    <geom class='tt-block-center'/>
                    <geom class='tt-block-right'/>
                </body>
                
                <body euler='0 0 1.57' pos='0 0 0.421'>
                    <geom class='tt-block-left'/>
                    <geom class='tt-block-center'/>
                    <geom class='tt-block-right'/>
                </body> -->
                
            </worldbody>
        </mujoco>
    """
)
class TumbleTower:
    pass


if __name__ == '__main__':
    from yarok import Platform, Injector
    from yarok.comm.worlds.empty_world import EmptyWorld
    from yarok.comm.components.ur5e.ur5e import UR5e
    from yarok.comm.components.robotiq_2f85.robotiq_2f85 import Robotiq2f85

    from math import pi

    @component(
        extends=EmptyWorld,
        components=[
            TumbleTower,
            UR5e,
            Robotiq2f85
        ],
        template="""
            <mujoco>
                <worldbody>
                    <tumble-tower name="tower"/>
                    
                    <body euler="0 0 1.57">
                        <ur5e name='arm'>
                           <robotiq-2f85 name="gripper" parent="ee_link"> 
                            </robotiq-2f85> 
                        </ur5e>
                    </body>
                </worldbody>
            </mujoco>
        """
    )
    class BlocksTowerTestWorld:
        pass


    class TumbleTheTowerBehaviour:

        def __init__(self, injector: Injector):
            self.arm: UR5e = injector.get('arm')
            self.pl: Platform = injector.get(Platform)

        def on_update(self):
            self.pl.wait_seconds(10)
            self.pl.wait(
                self.arm.move_xyz(xyz=[0.8, 0.1, 0.12], xyz_angles=[3.11, -1.6e-7, -pi / 2])
            )

            self.pl.wait(
                self.arm.move_xyz(xyz=[0.1, 0.8, 0.12], xyz_angles=[3.11, -1.6e-7, -pi / 2])
            )

            self.pl.wait_seconds(1000)


    Platform.create({
        'world': BlocksTowerTestWorld,
        'behaviour': TumbleTheTowerBehaviour,
        'defaults': {
            'plugins': []
        }
    }).run()



import time
import numpy as np
import cv2

from yarok import Platform


def normalize(img):
    if img.dtype == np.uint8:
        return img

    img1 = img - np.min(img)
    mx = np.max(img1)
    if -1e-6 < mx < 1e-6:
        return img1.astype(dtype=np.uint8)
    return ((img1 / mx) * 255).astype(dtype=np.uint8)

def torgb(img):
    if len(img.shape) == 3:
        return img

    return np.stack([img, img, img], axis=2)

class Cv2Inspector:

    def __init__(self, platform: Platform):
        self.components_manager = platform.manager
        self.platform = platform
        self.probe_interval = 0.5
        self.next_probe_ts = time.time() + self.probe_interval

    def inspect_component(self, comp):
        probe = self.components_manager.config(comp)['probe']

        if not callable(probe):
            return

        data = probe(comp['instance'])

        for name, datum in data.items():
            if type(datum) == np.ndarray:
                frame = datum.copy()
                frame = normalize(frame)
                frame = torgb(frame)
                minx_max_txt = "{:.3f}".format(np.min(datum)) + ' - ' + "{:.3f}".format(np.max(datum))
                frame = cv2.putText(frame, minx_max_txt, (20, 20),
                                    cv2.FONT_HERSHEY_SIMPLEX,
                                    0.5,
                                    (0, 255, 0),
                                    1,
                                    cv2.LINE_AA)
                frame_bgr = cv2.cvtColor(frame, cv2.COLOR_RGB2BGR)
                cv2.imshow(comp['name'] + '/' + name, frame_bgr)
                cv2.setWindowTitle(comp['name'] + '/' + name, comp['name'] + '/' + name)

    def step(self):
        """
            Calls probe
        """
        if time.time() > self.next_probe_ts:
            [
                self.inspect_component(comp)
                for component_id, comp in self.components_manager.components.items()
                if 'probe' in self.components_manager.config(comp)
            ]
            self.next_probe_ts = time.time() + self.probe_interval

            cv2.waitKey(1)



import yarok
from yarok import Platform, PlatformMJC, PlatformHW, component, ConfigBlock, Injector

@component(
    template="""
        <mujoco>
            <!-- possibly the physics solver -->
            <compiler angle="radian"  autolimits="true"/>
 <!--             <option impratio="10"/>

            <option timestep="0.01" solver="Newton" iterations="30" tolerance="1e-10" jacobian="auto" cone="pyramidal"/> -->
            <size nconmax="10000" njmax="10000"/>
            <visual>
                <!-- important for tactile sensors, to ensure the its camera frustum captures the close-up elastomer -->
                <map znear="0.001" zfar="100"/>
<!--                <quality shadowsize="2048"/> -->
            </visual>

            <asset>
                <!-- empty world -->
                <texture name="texplane" type="2d" builtin="checker" rgb1=".2 .3 .4" rgb2=".1 0.15 0.2"
                         width="512" height="512" mark="cross" markrgb=".8 .8 .8"/>    
                <material name="matplane" reflectance="0.3" texture="texplane" texrepeat="1 1" texuniform="true"/>
            </asset>

            <worldbody>
                <light directional="true" diffuse="1.0 1.0 1.0" specular="0.1 0.1 0.1" pos="0 0 5.0" dir="0 0 -1"/>
                <body name="floor">
                    <geom name="ground" type="plane" size="0 0 1" pos="0 0 0" quat="1 0 0 0" material="matplane"/>
                </body>
            </worldbody>

        </mujoco>
    """
)
class EmptyWorld:
    pass



class ConfigBlock:

    def __init__(self, config):
        self.config = config
        self._defaults = {}

    def defaults(self, defaults):
        self._defaults = defaults

    def __contains__(self, item):
        return item in self.config or item in self._defaults
    
    def __getitem__(self, item):

        def as_config_block(cfg, defaults=None):
            if type(cfg) == dict:
                block = ConfigBlock(cfg)
                if defaults is not None:
                    block.defaults(defaults)
                return block
            return cfg

        if item in self.config:
            return as_config_block(self.config[item],
                                   self._defaults[item] if item in self._defaults else {})
        if item in self._defaults:
            return as_config_block(self._defaults[item])

        raise KeyError()
        return None

    def is_config_block(self, key):
        return isinstance(self[key], ConfigBlock)



from yarok import ConfigBlock, Platform, component, interface
from yarok.platforms.mjc import InterfaceMJC

import numpy as np
import cv2
import os

import math
from time import time

from .utils.phong_render import PhongRender

gel_width, gel_height = 240, 320

__location__ = os.path.realpath(
    os.path.join(os.getcwd(), os.path.dirname(__file__)))


@interface(
    defaults={
        'light_sources': [
            {'position': [-0.866, 0.5, 0.25], 'color': (255, 82, 108), 'kd': 0.1, 'ks': 0.4},
            # red, bottom (108, 82, 255) kd 0.6
            {'position': [0.866, 0.5, 0.25], 'color': (153, 255, 120), 'kd': 0.1, 'ks': 0.4},
            # green, left (120, 255, 153)
            {'position': [0, -1, 0.25], 'color': (115, 130, 255), 'kd': 0.3, 'ks': 0.4},
            # blue, left (255, 130, 115)
        ],
        'background_img': cv2.cvtColor(
            cv2.imread(os.path.join(__location__, './assets/gel/digit_bg.png')),
            cv2.COLOR_BGR2RGB),
        'ka': 1.0,
        'texture_sigma': 0.00001,
        'px2m_ratio': 5.4347826087e-05,
        'elastomer_thickness': 1.0,
        'min_depth': 0.0
    }
)
class DigitInterfaceMJC:

    def __init__(self, interface: InterfaceMJC, config: ConfigBlock):
        self.interface = interface
        self.simulation = PhongRender(
            light_sources=config['light_sources'],
            background_img=config['background_img'],
            ka=config['ka'],
            texture_sigma=config['texture_sigma'],
            px2m_ratio=config['px2m_ratio'],
            elastomer_thickness=config['elastomer_thickness'],
            min_depth=config['min_depth']
        )
        self.last_update = 0
        self.depth = np.zeros((320, 240), np.float32)

    def read(self, shape=(320, 240)):
        t = time()
        if self.last_update > t - 1.0:
            return self.tactile
        self.last_update = t
        self.depth = self.interface.read_camera('digit0:camera', shape, depth='raw', rgb=False)
        self.tactile = self.simulation.generate(self.depth).astype(np.uint8)

        return self.tactile

    def read_depth(self):
        return self.depth


class DigitInterfaceHW:

    def __init__(self, config: ConfigBlock):
        self.cap = cv2.VideoCapture(config['cam_id'])
        if not self.cap.isOpened():
            raise Exception('Digit cam ' + str(config['cam_id']) + ' not found')

        self.fake_depth = np.zeros((240, 320), np.float32)

    def read(self):
        [self.cap.read() for _ in range(10)]  # skip frames in the buffer.
        ret, frame = self.cap.read()
        return frame

    def read_depth(self):
        return self.fake_depth


@component(
    tag="digit",
    defaults={
        'interface_mjc': DigitInterfaceMJC,
        'interface_hw': DigitInterfaceHW,
        'probe': lambda c: {'camera': c.read()},
    },
    template_path='digit.xml'
)
class Digit:

    def __init__(self):
        """
            DIGIT simulation as proposed in
            https://github.com/Rancho-zhao/Digit_PhongSim
        """
        pass

    def read(self):
        pass

    def read_depth(self):
        pass


if __name__ == '__main__':
    Platform.create({
        'world': Digit
    }).run()



import os
import shutil


def generate(template):
    cwd = os.getcwd()
    drafts_path = os.path.dirname(os.path.abspath(__file__))

    cwd_file = os.path.join(cwd, template + '.py')
    drafts_file = os.path.join(drafts_path, 'drafts', template + '.py')

    if os.path.exists(cwd_file):
        print(f'File {template}.py already exits.')
        return

    print(f'Generated {template}.py')
    shutil.copyfile(drafts_file, cwd_file)